import os
import json
from unittest import TestCase

from pelecanus import PelicanJson
from pelecanus.toolbox import backfill_append
from pelecanus.toolbox import new_json_from_path
from pelecanus.toolbox import find_value
from pelecanus.toolbox import count_key
from pelecanus.toolbox import generate_paths
from pelecanus.toolbox import generate_paths_to_key
from pelecanus.toolbox import reverse_result
from pelecanus.toolbox import get_path
from pelecanus.toolbox import get_nested_value
from pelecanus.toolbox import set_nested_value

# Fixture locations
current_dir = os.path.abspath(os.path.dirname(__file__))
fixture_dir = os.path.join(current_dir, 'fixtures')
# Actual datasets
data = os.path.join(fixture_dir, 'test_data.json')
book = os.path.join(fixture_dir, 'book.json')
ricketts = os.path.join(fixture_dir, 'ricketts.json')
monterrey = os.path.join(fixture_dir, 'monterrey.json')
pelecanus_occidentalis = os.path.join(fixture_dir,
                                      'pelecanus_occidentalis.json')


class TestBackfillAppend(TestCase):

    def setUp(self):
        self.test_list = list(range(10))

    def test_original_list_unmodified(self):
        new_list = backfill_append(self.test_list, 20, "TEST")
        self.assertNotEqual(self.test_list, new_list)

    def test_modify_regular_index(self):
        new_list = backfill_append(self.test_list, 5, "TEST")
        self.assertEqual(new_list[5], "TEST")

        new_list = backfill_append(self.test_list, 0, "TEST")
        self.assertEqual(new_list[0], "TEST")

    def test_with_empty_list(self):
        newlist = []
        test_list = backfill_append(newlist, 0, "TEST")
        self.assertEqual(test_list[0], "TEST")

    def test_backfill(self):
        none_list = [None for x in range(10)]
        new_list = backfill_append(self.test_list, 20, "TEST")
        self.assertEqual(new_list[10:20], none_list)
        self.assertEqual(new_list[20], "TEST")


class TestNewJsonFromPath(TestCase):

    def test_no_path(self):
        self.assertEqual(new_json_from_path([], "TEST"),
                         "TEST")

    def test_dicts(self):
        expected = {'one': {'two': {'three': {'four': {'five': "VALUE"}}}}}
        path = ['one', 'two', 'three', 'four', 'five']
        new_json = new_json_from_path(path, "VALUE")
        self.assertEqual(expected, new_json)

    def test_lists(self):
        expected = [None, [None, None, [None, None, None, [None, None,
                                                           None, None,
                                                           "VALUE"]]]]
        path = [1, 2, 3, 4]
        new_json = new_json_from_path(path, "VALUE")
        self.assertEqual(expected, new_json)

    def test_mixed(self):
        expected = {'one': {'two': [None, None, None, [{'four': "VALUE"}]]}}
        path = ['one', 'two', 3, 0, 'four']
        new_json = new_json_from_path(path, "VALUE")
        self.assertEqual(expected, new_json)


class TestFindValue(TestCase):

    def setUp(self):
        with open(ricketts, 'r') as f:
            self.ricketts = json.loads(f.read())
        with open(pelecanus_occidentalis, 'r') as f:
            self.pelecanus_occidentalis = json.loads(f.read())
        with open(data, 'r') as f:
            rdata = json.loads(f.read())
            self.item = rdata['items'][-1]

    def test_find_value(self):
        expected = ['query', 'normalized', 0, 'from']
        seek_value = 'Pelecanus_occidentalis'
        value_gen = find_value(self.pelecanus_occidentalis, seek_value)
        self.assertEqual(next(value_gen), expected)
        with self.assertRaises(StopIteration):
            next(value_gen)
        answer_template = ['query', 'pages', '1422396', 'images']
        answers = [answer_template + [n, 'ns'] for n in range(7)]
        for path in find_value(self.ricketts, 6):
            self.assertIn(path, answers)

    def test_find_value_inside_lists(self):
        paths = [['links', 'enclosure', 0, 'meta', 'program_guid'],
                 ['attributes', 'tags', 0],
                 ['attributes', 'guid']]
        for path in find_value(self.item, 'someGUID'):
            self.assertIn(path, paths)


class TestCountKey(TestCase):

    def setUp(self):
        with open(data, 'r') as f:
            rdata = json.loads(f.read())
            self.item = rdata['items'][-1]
        with open(ricketts, 'r') as f:
            self.ricketts = json.loads(f.read())

    def test_count_key(self):
        print("counting 'href'")
        self.assertEqual(count_key(self.item, 'href'), 11)
        print("counting 'extlinks'")
        self.assertEqual(count_key(self.ricketts, 'extlinks'), 2)
        print("counting '*'")
        self.assertEqual(count_key(self.ricketts, '*'), 10)
        print("counting 'title'")
        self.assertEqual(count_key(self.ricketts, 'title'), 8)


class TestGeneratePaths(TestCase):

    def setUp(self):
        with open(book, 'r') as f:
            self.book = json.loads(f.read())
        with open(data, 'r') as f:
            rdata = json.loads(f.read())
            self.item = rdata['items'][8]

    def test_generate_paths(self):
        book_paths = [['ISBN:9780804720687'],
                      ['ISBN:9780804720687', 'thumbnail_url'],
                      ['ISBN:9780804720687', 'bib_key'],
                      ['ISBN:9780804720687', 'preview_url'],
                      ['ISBN:9780804720687', 'info_url'],
                      ['ISBN:9780804720687', 'preview']]
        for path in generate_paths(self.book):
            self.assertIn(path, book_paths)

    def test_generate_paths_inside_list(self):
        test_pelican = PelicanJson(self.item)
        expected_paths = []
        for path in test_pelican.search_value(''):
            expected_paths.append(path)
        for path in generate_paths(self.item):
            value = get_nested_value(self.item, path)
            if value == '':
                self.assertIn(path, expected_paths)


class TestGeneratePathsToKey(TestCase):

    def setUp(self):
        with open(book, 'r') as f:
            self.book = json.loads(f.read())
        with open(data, 'r') as f:
            rdata = json.loads(f.read())
            self.item = rdata['items'][0]

    def test_generate_paths_to_key(self):
        book_template = ['query', 'pages', '1422396', 'extlinks']
        book_paths = [book_template + [n, "*"] for n in range(10)]
        for path in generate_paths_to_key(self.book, "*"):
            self.assertIn(path, book_paths)

    def test_generate_paths_to_key_inside_list(self):
        paths = [['links', 'profile', 0, 'type'],
                 ['links', 'enclosure', 0, 'type'],
                 ['links', 'schema', 0, 'type'],
                 ['links', 'documentation', 0, 'type'],
                 ['links', 'self', 0, 'type'],
                 ['links', 'edit-form', 0, 'type'],
                 ['links', 'extends', 0, 'type']]
        for path in generate_paths_to_key(self.item, 'type'):
            self.assertIn(path, paths)


class TestReverseResultDecorator(TestCase):

    def test_reverse_result(self):
        answer = "THIs is a A reversed string"

        def dummy_func1():
            return list(range(19, 0, -1))

        def dummy_func2():
            return answer[::-1]

        range_test = reverse_result(dummy_func1)
        string_test = reverse_result(dummy_func2)

        self.assertEqual(range_test(), list(dummy_func1())[::-1])
        self.assertEqual(string_test(), "".join(list(dummy_func2())[::-1]))

    def test_reverse_result_should_return_none(self):
        def dummy_func():
            return None
        reverser = reverse_result(dummy_func)
        self.assertEqual(reverser(), None)


class TestGetPath(TestCase):
    def setUp(self):
        with open(data, 'r') as f:
            rdata = json.loads(f.read())
            self.item = rdata['items'][4]

    def test_get_path(self):
        found_path1 = get_path(self.item, 'premiere_date')
        answer1 = ['links', 'enclosure', 0, 'meta', 'premiere_date']
        found_path2 = get_path(self.item, 'documentation')
        answer2 = ['links', 'documentation']
        self.assertEqual(found_path1, answer1)
        self.assertEqual(found_path2, answer2)
        self.assertEqual(get_path(self.item, 'NO PATH'),
                         None)


class TestGetNestedValue(TestCase):
    def setUp(self):
        with open(data, 'r') as f:
            rdata = json.loads(f.read())
            self.item = rdata['items'][-1]
        with open(monterrey, 'r') as f:
            self.monterrey = json.loads(f.read())

    def test_get_nested_value(self):
        answer = get_nested_value(self.item, ['attributes', 'tags', 2])
        self.assertEqual(answer, 'Cove')
        answers = [('gov.noaa.ncdc:C00822', ['results', 7, 'uid']),
                   ('gov.noaa.ncdc:C00040', ['results', 0, 'uid'])]
        for answer, path in answers:
            self.assertEqual(get_nested_value(self.monterrey, path),
                             answer)
        # Testing with a tuple, also should be allowed
        path = ('results', 9, 'uid')
        self.assertEqual(get_nested_value(self.monterrey, path),
                         'gov.noaa.ncdc:C00313')

    def test_get_nested_value_raises_empty_path(self):
        self.assertEqual(get_nested_value(self.monterrey, []),
                         None)


class TestSetNestedValue(TestCase):
    def setUp(self):
        with open(data, 'r') as f:
            rdata = json.loads(f.read())
            self.item = rdata['items'][4]
        with open(book, 'r') as f:
            self.book = json.loads(f.read())
        with open(pelecanus_occidentalis, 'r') as f:
            self.pelecanus_occidentalis = json.loads(f.read())

    def test_set_nested_value(self):
        new_path = ['ISBN:9780804720687', 'book_title']
        set_nested_value(self.book, new_path, 'Between Pacific Tides')
        self.assertEqual(get_nested_value(self.book, new_path),
                         'Between Pacific Tides')

        set_nested_value(self.pelecanus_occidentalis, ['href'], None)
        self.assertEqual(self.pelecanus_occidentalis['href'], None)

    def test_set_all_paths(self):
        test_pelican = PelicanJson(self.item)
        for path in test_pelican.paths():
            set_nested_value(self.item, path, "NEWVALUE")
        for path in test_pelican.paths():
            value = get_nested_value(self.item, path)
            self.assertEqual(value, "NEWVALUE")



import os
import json
import copy
from unittest import TestCase

from pelecanus import PelicanJson
from pelecanus.exceptions import BadPath
from pelecanus.exceptions import EmptyPath


# Fixture locations
current_dir = os.path.abspath(os.path.dirname(__file__))
fixture_dir = os.path.join(current_dir, 'fixtures')
# Actual datasets
data = os.path.join(fixture_dir, 'datadoc.json')
book = os.path.join(fixture_dir, 'book.json')
ricketts = os.path.join(fixture_dir, 'ricketts.json')
monterrey = os.path.join(fixture_dir, 'monterrey.json')
pelecanus_occidentalis = os.path.join(fixture_dir,
                                      'pelecanus_occidentalis.json')


class TestDictIntegrity(TestCase):

    def setUp(self):
        with open(data, 'r') as f:
            self.data = json.loads(f.read())

    def test_constructor(self):
        test_pelican = PelicanJson(self.data['items'][0])
        self.assertTrue(isinstance(test_pelican['attributes'],
                                   PelicanJson))
        self.assertTrue(isinstance(test_pelican['links'],
                                   PelicanJson))
        self.assertTrue(isinstance(test_pelican['attributes']['tags'],
                                   list))

    def test_convert(self):
        test_pelican = PelicanJson(self.data)
        self.assertEqual(test_pelican.convert(), self.data)
        self.assertTrue(isinstance(test_pelican.convert(),
                                   dict))

    def test_serialize(self):
        test_pelican = PelicanJson(self.data)
        self.assertEqual(json.loads(test_pelican.serialize()),
                         self.data)

    def general_test_enum_nestedval(self):
        test_pelican = PelicanJson(self.data)
        for path, value in test_pelican.enumerate():
            retrieved_value = test_pelican.get_nested_value(path)
            self.assertEqual(retrieved_value,
                             value)


class TestDunderMethods(TestCase):

    def setUp(self):
        with open(book, 'r') as f:
            self.book = json.loads(f.read())
        with open(ricketts, 'r') as f:
            self.ricketts = json.loads(f.read())
        with open(monterrey, 'r') as f:
            self.monterrey = json.loads(f.read())
        with open(pelecanus_occidentalis, 'r') as f:
            self.pelecanus_occidentalis = json.loads(f.read())

    def test_to_string(self):
        test_pelican = PelicanJson(self.ricketts)
        self.assertNotIn("<PelicanJson:", str(test_pelican))

    def test_update_from_list(self):
        test_pelican = PelicanJson(self.ricketts)
        values = [{'*': '//www.worldcat.org/identities/lccn-n79-055298'},
                  {'*': 'http://edricketts.stanford.edu/'},
                  {'*': 'http://holisticbiology.stanford.edu/philosophy.html'}]
        extlinks = self.ricketts['query']['pages']['1422396']['extlinks']
        first, *_, second_last, last = test_pelican._update_from_list(extlinks)
        self.assertEqual(first, PelicanJson(values[0]))
        self.assertEqual(second_last, PelicanJson(values[1]))
        self.assertEqual(last, PelicanJson(values[2]))

    def test_contains(self):
        test_pelican = PelicanJson(self.book)
        self.assertIn('ISBN:9780804720687', test_pelican)
        self.assertIn('preview', test_pelican)
        self.assertIn('bib_key', test_pelican)
        test_pelican = PelicanJson(self.monterrey)
        for key in test_pelican.keys():
            self.assertIn(key, test_pelican)
        self.assertFalse(test_pelican.__contains__('NADA'))

    def test_get_item(self):
        query_continue = self.ricketts['query-continue']
        test_pelican = PelicanJson(self.ricketts)
        query_pelican = PelicanJson(query_continue)
        self.assertEqual(test_pelican.get('query-continue'),
                         query_pelican)
        self.assertEqual(test_pelican.get('NO-KEY'),
                         None)

    def test_set_item(self):
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        test_pelican['query'] = ['some', 'new' 'vals']
        self.assertEqual(test_pelican['query'], ['some', 'new' 'vals'])

    def test_set_item_dictionary(self):
        test_pelican = PelicanJson(self.book)
        newdict = {'somekey': 'somevalue',
                   'somenested': {'somenestedkey': 'somenestedvalue'}}
        expected = PelicanJson(newdict)
        test_pelican['TEST'] = newdict
        self.assertTrue(isinstance(test_pelican, PelicanJson))
        self.assertEqual(test_pelican['TEST'], expected)

    def test_set_item_list(self):
        test_pelican = PelicanJson(self.book)
        newlist = ['item1', 'item2', {'somenestedkey': 'somenestedvalue'}]
        expected = PelicanJson({'newlist': newlist})
        test_pelican['newlist'] = newlist
        self.assertEqual(test_pelican['newlist'], expected['newlist'])

    def test_delete_item(self):
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        test_pelican['query'] = ['some', 'new' 'vals']
        self.assertEqual(test_pelican['query'], ['some', 'new' 'vals'])
        del test_pelican['query']
        self.assertEqual(test_pelican.get('query'),
                         None)
        with self.assertRaises(KeyError):
            del test_pelican['none']

    def test_len(self):
        keys = {'query', 'normalized', 'to', 'from', 'pages',
                '1266004', 'ns', 'title', 'pageid'}
        length = len(keys)
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        self.assertEqual(len(test_pelican), length)
        self.assertEqual(set(test_pelican.keys()), keys)
        self.assertEqual(len(test_pelican), len(list(test_pelican.keys())),
                         len(keys))
        test_pelican['new'] = 'VALUE'
        self.assertEqual(len(test_pelican), length + 1)

    def test_iter(self):
        pkeys = {'query', 'normalized', 'to', 'from', 'pages',
                 '1266004', 'ns', 'title', 'pageid'}
        bkeys = {'thumbnail_url', 'preview_url', 'bib_key',
                 'info_url', 'ISBN:9780804720687', 'preview'}
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        test_book = PelicanJson(self.book)
        self.assertEqual(set(test_pelican.keys()), pkeys)
        self.assertEqual(set(test_book.keys()), bkeys)


class TestPelicanMethods(TestCase):

    def setUp(self):
        with open(data, 'r') as f:
            rdata = json.loads(f.read())
            self.item = rdata['items'][-1]
        with open(book, 'r') as f:
            self.book = json.loads(f.read())
        with open(ricketts, 'r') as f:
            self.ricketts = json.loads(f.read())
        with open(monterrey, 'r') as f:
            self.monterrey = json.loads(f.read())
        with open(pelecanus_occidentalis, 'r') as f:
            self.pelecanus_occidentalis = json.loads(f.read())

    def test_items(self):
        answer_dict = {'ns': 0,
                       'title': 'Pelecanus occidentalis',
                       'pageid': 1266004,
                       'to': 'Pelecanus occidentalis',
                       'from': 'Pelecanus_occidentalis'}
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        test_book = PelicanJson(self.book)
        self.assertEqual(len(list(test_book.keys())),
                         len(list(test_book.items())))
        self.assertEqual(len(list(test_pelican.keys())),
                         len(list(test_pelican.items())))
        for key, value in test_pelican.items():
            if key in answer_dict:
                self.assertEqual(answer_dict[key], value)

    def test_enumerate(self):
        book_url = 'https://openlibrary.org/books/OL7928788M/'
        book_url += 'Between_Pacific_Tides'
        test_book_enums = [
            (['ISBN:9780804720687', 'thumbnail_url'],
             'https://covers.openlibrary.org/b/id/577352-S.jpg'),
            (['ISBN:9780804720687', 'bib_key'],
             'ISBN:9780804720687'),
            (['ISBN:9780804720687', 'preview_url'], book_url),
            (['ISBN:9780804720687', 'info_url'], book_url),
            (['ISBN:9780804720687', 'preview'], 'noview')]
        test_book = PelicanJson(self.book)
        for path, value in test_book.enumerate():
            self.assertIn((path, value), test_book_enums)

    def test_paths(self):
        test_book = PelicanJson(self.book)
        book_paths = [['ISBN:9780804720687', 'thumbnail_url'],
                      ['ISBN:9780804720687', 'bib_key'],
                      ['ISBN:9780804720687', 'preview_url'],
                      ['ISBN:9780804720687', 'info_url'],
                      ['ISBN:9780804720687', 'preview']]
        for item in test_book.paths():
            self.assertIn(item, book_paths)

    def test_keys(self):
        pkeys = {'query', 'normalized', 'to', 'from', 'pages',
                 '1266004', 'ns', 'title', 'pageid'}
        bkeys = {'thumbnail_url', 'preview_url', 'bib_key',
                 'info_url', 'ISBN:9780804720687', 'preview'}
        test_book = PelicanJson(self.book)
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        self.assertEqual(set(test_pelican.keys()), pkeys)
        self.assertEqual(set(test_book.keys()), bkeys)
        self.assertEqual(len(test_pelican), len(set(test_pelican.keys())),
                         len(pkeys))

    def test_keys_flat(self):
        test_book = PelicanJson(self.book)
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        assert set(test_book.keys(flat=True)) == {'ISBN:9780804720687'}
        assert set(test_pelican.keys(flat=True)) == {"query"}

    def test_values(self):
        pvalues = {'Pelecanus occidentalis', 'Pelecanus_occidentalis',
                   0, 'Pelecanus occidentalis', 1266004}
        book_url = 'https://openlibrary.org/books/OL7928788M/'
        book_url += 'Between_Pacific_Tides'
        monty_uid = 'gov.noaa.ncdc:C00345'
        rval = 'Ed_Ricketts'
        rval2 = 'File:Pacific Biological Laboratories.JPG'
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        test_book = PelicanJson(self.book)
        test_monty = PelicanJson(self.monterrey)
        test_rickettsi = PelicanJson(self.ricketts)
        self.assertEqual(pvalues, set(test_pelican.values()))
        self.assertIn(book_url, set(test_book.values()))
        self.assertIn(monty_uid, set(test_monty.values()))
        self.assertIn(rval, set(test_rickettsi.values()))
        self.assertIn(rval2, set(test_rickettsi.values()))

    def test_count_key(self):
        test_pelican = PelicanJson(self.item)
        self.assertEqual(test_pelican.count_key('href'),
                         11)
        test_rickettsi = PelicanJson(self.ricketts)
        self.assertEqual(test_rickettsi.count_key('extlinks'),
                         2)
        self.assertEqual(test_rickettsi.count_key('*'),
                         10)
        self.assertEqual(test_rickettsi.count_key('title'),
                         8)

    def test_create_path_totally_new_path(self):
        test_rickettsi = PelicanJson(self.ricketts)
        path = ['new', 'path', 'in', 1, 'object']
        with self.assertRaises(KeyError):
            test_rickettsi.get_nested_value(path)
        test_rickettsi.create_path(path, "TEST VALUE")
        self.assertEqual(test_rickettsi.get_nested_value(path), "TEST VALUE")

    def test_create_path_in_dict(self):
        test_rickettsi = PelicanJson(self.ricketts)
        overwrite_path = ['query-continue', 'extlinks',
                          'eloffset', 'newdict-with-key']
        msg = "Previous value overwritten with dictionary"
        test_rickettsi.create_path(overwrite_path, msg)
        self.assertEqual(test_rickettsi.get_nested_value(overwrite_path),
                         msg)
        newdict = test_rickettsi.get_nested_value(overwrite_path[:-1])
        self.assertTrue(isinstance(newdict, PelicanJson))

        paths = [['query-continue', 'extlinks', 'newkey1'],
                 ['query-continue', 'newkey1', 'newkey2'],
                 ['newkey1', 'newkey2', 'newkey3']]
        for path in paths:
            with self.assertRaises(KeyError):
                test_rickettsi.get_nested_value(path)
            test_rickettsi.create_path(path, "NEWVALUE")
            self.assertEqual(test_rickettsi.get_nested_value(path),
                             "NEWVALUE")

    def test_create_path_new_object_inside_list(self):
        test_rickettsi = PelicanJson(self.ricketts)
        paths = [['query', 'pages', '1422396', 'images', 7, 'title'],
                 ['query', 'normalized', 10, 'NEW']]
        check_for_none = ['query', 'normalized', 5]
        test_rickettsi.create_path(paths[0], "VALUE APPENDED TO LIST")
        test_rickettsi.create_path(paths[1], "VALUE in LIST with BACKFILL")
        self.assertEqual(test_rickettsi.get_nested_value(paths[0]),
                         "VALUE APPENDED TO LIST")
        self.assertEqual(test_rickettsi.get_nested_value(paths[1]),
                         "VALUE in LIST with BACKFILL")
        self.assertEqual(test_rickettsi.get_nested_value(check_for_none),
                         None)
        self.assertEqual(len(test_rickettsi.get_nested_value(['query',
                                                              'normalized'])),
                         11)

    def test_create_path_add_item_to_list(self):
        test_item = PelicanJson(self.item)
        paths = [['attributes', 'tags', 2],
                 ['attributes', 'tags', 5]]
        check_for_none = paths[1][:-1]
        check_for_none.append(4)
        test_item.create_path(paths[0], "New value inside list")
        self.assertEqual(test_item.get_nested_value(paths[0]),
                         "New value inside list")
        test_item.create_path(paths[1], "New value inside list with None")
        self.assertEqual(test_item.get_nested_value(paths[1]),
                         "New value inside list with None")
        self.assertEqual(test_item.get_nested_value(check_for_none),
                         None)

    def test_create_path_raise_badpath(self):
        test_rickettsi = PelicanJson(self.ricketts)
        bad_path = [[4, 'query', 'normalized', 0, 'from'],
                    [('some', 'tuple'), 'query'],
                    [{'some': 'dict'}, 'query']]
        with self.assertRaises(BadPath):
            test_rickettsi.create_path(bad_path[0], "VALUE")
            test_rickettsi.create_path(bad_path[1], "VALUE")
            test_rickettsi.create_path(bad_path[2], "VALUE")

    def test_create_path_raise_indexerror(self):
        test_rickettsi = PelicanJson(self.ricketts)
        bad_path = ['query', 'normalized', 'badkey', 'from']
        with self.assertRaises(IndexError):
            test_rickettsi.create_path(bad_path, "VALUE")

    def test_searchkey(self):
        test_rickettsi = PelicanJson(self.ricketts)
        paths = [['query', 'pages', '1422396', 'extlinks'] + [n, '*']
                 for n in range(10)]
        for item in test_rickettsi.search_key('*'):
            self.assertIn(item, paths)
        for key in test_rickettsi.keys():
            self.assertTrue(next(test_rickettsi.search_key(key)))
        pelican_item = PelicanJson(self.item)
        for key in pelican_item.keys():
            self.assertTrue(next(pelican_item.search_key(key)))

    def test_searchvalue(self):
        test_monty = PelicanJson(self.monterrey)
        values = list(test_monty.values())
        self.assertEqual(len(values), 63)
        answers = [('gov.noaa.ncdc:C00822', ['results', 7, 'uid']),
                   ('gov.noaa.ncdc:C00040', ['results', 0, 'uid'])]
        for item, answer in answers:
            self.assertEqual(next(test_monty.search_value(item)),
                             answer)
        self.assertEqual(list(test_monty.search_value('2014-08-25')),
                         [['results', 1, 'maxdate'],
                          ['results', 3, 'maxdate']])
        npr_api_tag = [['attributes', 'tags', 0],
                       ['items', 0, 'attributes', 'tags', 0]]
        pelican_item = PelicanJson(self.item)
        for path in pelican_item.search_value('npr_api'):
            self.assertIn(path, npr_api_tag)

    def test_pluck(self):
        answer = PelicanJson({'to': 'Pelecanus occidentalis',
                              'from': 'Pelecanus_occidentalis'})
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        self.assertEqual(answer,
                         next(test_pelican.pluck('to',
                                                 'Pelecanus occidentalis')))
        self.assertEqual(answer,
                         next(test_pelican.pluck('from',
                                                 'Pelecanus_occidentalis')))
        pelican_item = PelicanJson(self.item)
        self.assertEqual(pelican_item['attributes'],
                         next(pelican_item.pluck('byline', 'Emily Reddy')))

    def test_get_nested_value(self):
        pelican_item = PelicanJson(self.item)
        answer = pelican_item.get_nested_value(['attributes', 'tags', 0])
        self.assertEqual(answer, 'npr_api')
        test_monty = PelicanJson(self.monterrey)
        answers = [('gov.noaa.ncdc:C00822', ['results', 7, 'uid']),
                   ('gov.noaa.ncdc:C00040', ['results', 0, 'uid'])]
        for answer, path in answers:
            self.assertEqual(test_monty.get_nested_value(path),
                             answer)
        # Testing with a tuple, also should be allowed
        path = ('results', 9, 'uid')
        self.assertEqual(test_monty.get_nested_value(path),
                         'gov.noaa.ncdc:C00313')

    def test_get_nested_value_raises_empty_path(self):
        test_pelican = PelicanJson(self.monterrey)
        with self.assertRaises(EmptyPath):
            test_pelican.get_nested_value([])

    def test_get_nested_value_raises_bad_path(self):
        test_pelican = PelicanJson(self.monterrey)
        # Try a string
        with self.assertRaises(BadPath):
            test_pelican.get_nested_value("STRING")
        # Howsbout a dict?
        somedict = {'results': 'value', '9': 'value2'}
        with self.assertRaises(BadPath):
            test_pelican.get_nested_value(somedict)
        # What happens with an integer?
        with self.assertRaises(BadPath):
            test_pelican.get_nested_value(8)
        # ...and a set?
        with self.assertRaises(BadPath):
            test_pelican.get_nested_value({'results', 8})

    def test_set_nested_value(self):
        new_path = ['ISBN:9780804720687', 'book_title']
        test_book = PelicanJson(self.book)
        test_book.set_nested_value(new_path, 'Between Pacific Tides')
        self.assertEqual(test_book.get_nested_value(new_path),
                         'Between Pacific Tides')

        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        values = []
        self.assertEqual(len(list(test_pelican.values())), 5)
        for path, value in test_pelican.enumerate():
            values.append(value)
            test_pelican.set_nested_value(path, None)
        self.assertEqual(len(set(test_pelican.values())), 1)
        for path in test_pelican.search_value(None):
            test_pelican.set_nested_value(path, values.pop())
        self.assertEqual(len(list(test_pelican.values())), 5)

        pelican_item = PelicanJson(self.item)
        pelican_item.set_nested_value(['href'], None)
        self.assertEqual(pelican_item['href'], None)

    def test_set_nested_value_raises_error(self):
        # Wish to raise IndexError, KeyError, TypeError
        test_rickettsi = PelicanJson(self.ricketts)
        key_error_paths = [['unknownKey', 'unknownKey2'],
                           ['query-continue', 'unknownKey']]
        # KeyErrors
        with self.assertRaises(KeyError):
            for path in key_error_paths:
                test_rickettsi.set_nested_value(path, "Shouldn't Work")

        # TypeErrors
        type_error_path = ['query-continue', 'extlinks',
                           'eloffset', 'newdict-with-key']
        with self.assertRaises(TypeError):
            test_rickettsi.set_nested_value(type_error_path,
                                            "Shouldn't Work")

        # IndexErrors
        index_error_paths = [['query', 'normalized', 1, 'from'],
                             ['query', 'normalized', 1, 'to']]
        with self.assertRaises(IndexError):
            for path in index_error_paths:
                test_rickettsi.set_nested_value(path, "Shouldn't Work")

    # Attempt to set paths that previously raised: IndexError, KeyError,
    # TypeError
    def test_set_nested_value_force_key_error(self):
        test_rickettsi = PelicanJson(self.ricketts)
        success_msg = "Should now work"
        # KeyErrors overidden
        key_error_paths = [['unknownKey', 'unknownKey2'],
                           ['query-continue', 'unknownKey']]
        for path in key_error_paths:
            test_rickettsi.set_nested_value(path,
                                            success_msg,
                                            force=True)
            self.assertEqual(test_rickettsi.get_nested_value(path),
                             success_msg)

    def test_set_nested_value_force_type_error(self):
        test_rickettsi = PelicanJson(self.ricketts)
        success_msg = "Should now work"
        # TypeErrors overridden: path created
        type_error_path = ['query-continue', 'extlinks',
                           'eloffset', 'newdict-with-key']
        test_rickettsi.set_nested_value(type_error_path,
                                        success_msg,
                                        force=True)
        self.assertEqual(test_rickettsi.get_nested_value(type_error_path),
                         success_msg)

    def test_set_nested_value_force_previous_index_error(self):
        test_rickettsi = PelicanJson(self.ricketts)
        success_msg = "Should now work"
        # IndexErrors overridden: paths created
        index_error_paths = [['query', 'normalized', 1, 'from'],
                             ['query', 'normalized', 1, 'to']]
        for path in index_error_paths:
            test_rickettsi.set_nested_value(path, success_msg, force=True)
            self.assertEqual(test_rickettsi.get_nested_value(path),
                             success_msg)

    def test_set_nested_value_force_add_to_list(self):
        path = ['attributes', 'tags', 4]
        test_pelican = PelicanJson(self.item)
        test_pelican.set_nested_value(path, 'New Tag', force=True)
        new_tag = test_pelican.get_nested_value(path)
        self.assertEqual(new_tag, 'New Tag')
        none_placeholder = ['attributes', 'tags', 3]
        self.assertEqual(test_pelican.get_nested_value(none_placeholder),
                         None)

    def test_safe_get_nested_value(self):
        test_pelican = PelicanJson(self.monterrey)
        assert test_pelican.safe_get_nested_value(["STRING"]) is None
        badpath = ('results', 'value', '9')
        assert test_pelican.safe_get_nested_value(badpath,
                                                  default="la") == "la"
        assert test_pelican.safe_get_nested_value([8], default=0) == 0
        tp = (1, 2)
        assert test_pelican.safe_get_nested_value(['results', 1000],
                                                  default=tp) == tp

        assert test_pelican.safe_get_nested_value(["results", 10000],
                                                  default="test") == "test"
        assert test_pelican.safe_get_nested_value(["results", "key"],
                                                  default="test") == "test"

        with self.assertRaises(EmptyPath):
            test_pelican.safe_get_nested_value([])
        with self.assertRaises(BadPath):
            test_pelican.safe_get_nested_value({'results', 8})

    def test_find_and_replace(self):
        test_pelican = PelicanJson(self.pelecanus_occidentalis)
        test_pelican.find_and_replace('Pelecanus occidentalis',
                                      'Brown Pelican')
        replace_paths = [['query', 'normalized', 0, 'to'],
                         ['query', 'pages', '1266004', 'title']]
        for path in test_pelican.search_value('Brown Pelican'):
            self.assertIn(path, replace_paths)



"""
Pelecanus: an application for navigating and editing
nested JSON.

Named 'pelecanus' after Pelecanus occidentalis, the
brown Pelican of California and the Eastern Pacific:

http://www.nps.gov/chis/naturescience/brown-pelican.htm

which is a wonderful bird and which deserves to have something
named after it and because I got tired of writing "NestedJson".
"""
from .pelicanjson import PelicanJson  # noqa

__version__ = '0.5.2'



"""JSON is a recursive data structure, so this module has been created in
order to (hopefully) make parsing JSON easier to deal with.

Often, it's necessary to explore a JSON object without knowing precisely where
things are (in the case of Hypermedia, for example). By creating a recursive
data structure, we can facilitate such tasks as retrieving key-value pairs,
iterating through the data structure, and searching for elements in the data
structure.

There are limitations, however: no JSON structure that has keys deeper
than Python's recursion limit (default: 1000 stack frames) will work.

In addition, a JSON object that is a top-level array won't work, but I don't
actually think that's allowed, per JSON spec.
"""
import copy
import json
import collections

from .toolbox import backfill_append
from .toolbox import new_json_from_path

from .exceptions import BadPath
from .exceptions import EmptyPath


class PelicanJson(collections.MutableMapping):
    """PelicanJson objects are nested JSON objects that provide a few
    methods to make it easier to navigate and edit nested JSON objects.

    At the moment, this object can only handle JSON objects at the top-level,
    not arrays.

    To create an object, you can pass the constructor a Python
    dictionary created from a JSON dump::

       >>> content = {'links': {'attributes': [{'href': 'somelink'}]}}
       >>> pelican = PelicanJson(content)

    Once you have PelicanJson object, you can find all the nested paths and
    the values located at those paths::

       >>> for item in pelican.enumerate():
       ...   print(item)
       (['links', alternate', 0, 'href'], 'somelink')

    You can also get back the value from a nested path::

       >>> pelican.get_nested_value(['links', alternate', 0, 'href'])
       'somelink'

    If you want to change a nested value, you can use the `set_nested_value`
    command to do so::

       >>> pelican.set_nested_value(['links', alternate', 0, 'href'],
       ... 'newvalue')
       >>> pelican.get_nested_value(['links', alternate', 0, 'href'])
       'newvalue'

    A PelicanJson object is a modified version of a Python dictionary, so
    you can use all of the normal dictionary methods, but it will mostly return
    nested results (which means you will often get duplicate 'keys')::

       >>> list(pelican.keys())
       ['links', 'attributes', 'href']

    Other useful methods include `convert` and `serialize` for turning the
    object back into a plain Python dictionary and for returning a JSON dump,
    respectively::

       >>> pelican.convert() == content
       True
       >>> import json
       >>> json.loads(pelican.serialize()) == content
       True

    You can also `searchkey` and `searchvalue` in order to find all the paths
    that leads to keys or values you are searching for.
    """
    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.update(dict(*args, **kwargs))
        for key, value in self.store.items():
            # __setitem__ does the heavy-lifting here
            self.store[key] = value

    def __getitem__(self, key):
        return self.store[key]

    def __setitem__(self, key, value):
        if isinstance(value, dict):
            self.store[key] = PelicanJson(value)
        elif isinstance(value, list):
            self.store[key] = self._update_from_list(value)
        else:
            self.store[key] = value

    def __delitem__(self, key):
        del self.store[key]

    def _update_from_list(self, somelist):
        """Used to parse list objects for nested dictionaries and turn
        those internal dictionaries into PelicanJson objects.
        """
        temp_list = []
        for item in somelist:
            if isinstance(item, dict):
                temp_list.append(PelicanJson(item))
            elif isinstance(item, list):
                temp_list.append(self._update_from_list(item))
            else:
                temp_list.append(item)
        return temp_list

    def __len__(self):
        """Counts all keys and subkeys nested in the object.
        """
        return sum(1 for k in iter(self))

    def __contains__(self, searchkey):
        """Returns True if key is somewhere inside the object.
        """
        for key in iter(self):
            if key == searchkey:
                return True
        return False

    def __iter__(self):
        """Iterates through the entire tree and yields all nested keys.
        """
        for k, v in self.store.items():
            yield k
            if isinstance(v, PelicanJson):
                yield from iter(v)
            elif isinstance(v, list):
                for item in v:
                    if isinstance(item, PelicanJson):
                        yield from iter(item)

    def __repr__(self):
        return "<PelicanJson: {}>".format(str(self.store))

    def __str__(self):
        return str(self.convert())

    def items(self, path=None):
        """Yields path-value pairs from throughout the entire tree.
        """
        for k, v in self.store.items():
            yield k, v
            if isinstance(v, PelicanJson):
                yield from v.items()
            elif isinstance(v, list):
                for list_item in v:
                    if isinstance(list_item, PelicanJson):
                        yield from list_item.items()

    def enumerate(self, path=None):
        """Iterate through the PelicanJson object yielding 1) the full path to
        each value and 2) the value itself at that path.
        """
        if path is None:
            path = []
        for k, v in self.store.items():
            current_path = path[:]
            current_path.append(k)

            if isinstance(v, PelicanJson):
                yield from v.enumerate(path=current_path)
            elif isinstance(v, list):
                for idx, list_item in enumerate(v):
                    list_path = current_path[:]
                    list_path.append(idx)
                    if isinstance(list_item, PelicanJson):
                        yield from list_item.enumerate(path=list_path)
                    else:
                        yield list_path, list_item
            else:
                yield current_path, v

    def paths(self):
        """Uses enumerate to yield paths only
        """
        for path, _ in self.enumerate():
            yield path

    def keys(self, flat=False):
        """Generator that iterates through the keys of the nested object.

        kwargs:
           `flat` (bool): whether to return only top-level keys.
        """
        if flat:
            yield from self.store.keys()
        else:
            yield from iter(self)

    def values(self):
        """Generator that returns values-only for the object.
        """
        yield from (v for k, v in self.enumerate())

    def convert(self):
        """Converts the object back to a native Python object (a nested dictionary)
        that is equal to object passed in or, if modified, the dict version of
        self.store.
        """
        data = {}
        for k, v in self.store.items():
            if isinstance(v, PelicanJson):
                data[k] = v.convert()
            elif isinstance(v, list):
                temp_list = []
                for list_item in v:
                    if isinstance(list_item, PelicanJson):
                        temp_list.append(list_item.convert())
                    else:
                        temp_list.append(list_item)
                listcopy = copy.deepcopy(temp_list)
                data[k] = listcopy
            else:
                data[k] = v
        return data

    def serialize(self):
        """Returns JSON serialization of the object.
        """
        return json.dumps(self.convert())

    def count_key(self, key):
        """Returns a sum of the number of times a particular key appears in the object.
        """
        return sum(1 for k, v in self.items() if k == key)

    def create_path(self, path, newvalue):
        """Creates a new `path` set to `newvalue`.
        """
        def test_path(path):
            try:
                self.get_nested_value(path)
                return True
            except (IndexError, KeyError, TypeError):
                return False
        # The first element in the path needs to be a string so we can
        # add it as a key to self.store. We raise KeyError if otherwise
        if not isinstance(path[0], str) or len(path) == 0:
            errmsg = "New PelicanJson path must start with an acceptable key"
            errmsg += " (it must be a string). Bad path: {}"
            raise BadPath(errmsg.format(str(path)))

        # Now, let's take this path and figure out how much of it is already
        # present in the object.
        keys_present = path[:]
        keys_missing = collections.deque()
        while keys_present and not test_path(keys_present):
            keys_missing.appendleft(keys_present.pop())

        if keys_present:
            # The following object could be any value really
            edit_object = self.get_nested_value(keys_present)
            if isinstance(edit_object, list):
                # if list, we try to insert at the proper (missing) index
                index, *rest = keys_missing
                if not isinstance(index, int):
                    errmsg = "Check path. List index must be integer: {}."
                    raise IndexError(errmsg.format(index))
                if rest:
                    new_object = new_json_from_path(rest, newvalue)
                    new_object = PelicanJson(new_object)
                else:
                    new_object = newvalue
                edited_list = backfill_append(edit_object, index, new_object)
                self.set_nested_value(keys_present, edited_list)
            elif isinstance(edit_object, PelicanJson):
                new_object = new_json_from_path(keys_missing, newvalue)
                edit_object.update(new_object)
            else:
                # This is the case where we are overwriting some random value
                new_object = new_json_from_path(keys_missing, newvalue)
                self.set_nested_value(keys_present, new_object)
        else:
            # No keys_present: top-level object
            new_object = new_json_from_path(keys_missing, newvalue)
            self.update(new_object)

        return self

    def search_key(self, searchkey, path=None):
        """Generator that returns the (various) paths for a particular key
        """
        if path is None:
            path = []
        for k, v in self.store.items():
            current_path = path[:]
            current_path.append(k)
            if k == searchkey:
                yield current_path
            if isinstance(v, PelicanJson):
                yield from v.search_key(searchkey,
                                        path=current_path)
            elif isinstance(v, list):
                for idx, list_item in enumerate(v):
                    list_path = current_path[:]
                    list_path.append(idx)
                    if isinstance(list_item, PelicanJson):
                        yield from list_item.search_key(searchkey,
                                                        path=list_path)

    def search_value(self, searchval, path=None):
        """Generator that returns the (various) paths for a particular value
        """
        if path is None:
            path = []
        for k, v in self.store.items():
            current_path = path[:]
            current_path.append(k)
            if v == searchval:
                yield current_path

            if isinstance(v, PelicanJson):
                yield from v.search_value(searchval,
                                          path=current_path)
            elif isinstance(v, list):
                for idx, list_item in enumerate(v):
                    list_path = current_path[:]
                    list_path.append(idx)
                    if isinstance(list_item, PelicanJson):
                        yield from list_item.search_value(searchval,
                                                          path=list_path)
                    elif list_item == searchval:
                        yield list_path

    def pluck(self, key, value):
        """Returns the _parent_ object that contains a particular key-value pair
        """
        for path in self.search_key(key):
            if self.get_nested_value(path) == value:
                if len(path) > 1:
                    path = path[:-1]
                    yield self.get_nested_value(path)
                else:
                    yield self

    def get_nested_value(self, path):
        """Retrieves nested value at the end of a path.

        Raises either IndexError or KeyError if any element of the path is
        missing.
        """
        def valgetter(data, keys):
            if len(keys) <= 1:
                return data[keys[0]]
            else:
                key, *keys = keys
                return valgetter(data[key], keys)

        if isinstance(path, list) or isinstance(path, tuple):
            if len(path) > 0:
                return valgetter(self.store, path)
            else:
                raise EmptyPath("Path must have at least one element.")
        else:
            errmsg = "Path passed in is not a list or a tuple"
            raise BadPath(errmsg.format(str(path)))

    def set_nested_value(self, path, newvalue, force=False):
        """Sets a nested_value to a new value using the path provided.
        Path must already exist for path to be set.
        """
        *keys, last_key = path
        if len(keys) > 0:
            try:
                editable = self.get_nested_value(keys)
                editable[last_key] = newvalue
            except (IndexError, KeyError, TypeError) as e:
                if force:
                    self.create_path(path, newvalue)
                else:
                    raise e
        else:
            key, *_ = path
            self.store[key] = newvalue

    def safe_get_nested_value(self, path, default=None):
        """Retrieves nested value at the end of a path. Returns `default`
        if path doesn't return a value.

        kwargs:

          default: if supplied, returns this instead of IndexError/KeyError
        """
        def valgetter(data, keys):
            if len(keys) <= 1:
                return data[keys[0]]
            else:
                key, *keys = keys
                return valgetter(data[key], keys)

        if isinstance(path, list) or isinstance(path, tuple):
            if len(path) > 0:
                try:
                    return valgetter(self.store, path)
                except (KeyError, IndexError, TypeError):
                    return default
            else:
                raise EmptyPath("Path must have at least one element.")
        else:
            errmsg = "Path passed in is not a list or a tuple"
            raise BadPath(errmsg.format(str(path)))

    def find_and_replace(self, matchval, replaceval):
        """Will replace all matched values with the replacement value
        passed in and will yield the paths associated with the
        changed values.
        """
        for path in self.search_value(matchval):
            self.set_nested_value(path, replaceval)



"""
Pelecanus
=========

A Python3 application for navigating and editing nested JSON objects. For more
information, please see README.md (on github).


Name
----
Named 'pelecanus' after Pelecanus occidentalis, the
brown Pelican of California and the Eastern Pacific:

http://www.nps.gov/chis/naturescience/brown-pelican.htm

which is a wonderful bird and which deserves to have something
named after it and because I got tired of writing "NestedJson".

More Info
---

See github project for more detailed usage instructions.
"""

from setuptools import setup
import os
import codecs
import re

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    # intentionally *not* adding an encoding option to open
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='pelecanus',
    packages=['pelecanus'],
    version=find_version('pelecanus', '__init__.py'),
    description='Python3 application for navigating and editing nested JSON',
    author='Erik Aker',
    author_email="eraker@gmail.org",
    url="https://github.com/pellagic-puffbomb/pelecanus",
    license="GNU General Public License v2",
    download_url="https://github.com/pellagic-puffbomb/pelecanus.git",
    keywords=["json", "hateoas"],
    tests_require=['pytest',
                   'coverage',
                   "flake8",
                   "pytest_cov",
                   "pytest-pep8"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries"
    ],
    install_requires=[],
    long_description=__doc__
)
